from pydoc import describe
import numpy as np
import json
from scipy import linalg
from tqdm.auto import tqdm

class Molecules:
   def __init__(self, nmuc=None, coordmol=None, unitcell=None, 
   supercell=None, unique=None, uniqinter=None,
   javg=None, sigma=None, nrepeat=None,
   iseed=None, invtau=None, temp=None, 
   lattice_file=None, params_file=None, results={}):
      """
      The molecules object represent a group of molecules. It has the relevant information regarding 
      the calculation of charge mobility such as the transfer integral of the pairs, the standard 
      deviation and structural information about the system.

      Args:
          nmuc (int, optional): Number of molcules in the unit cell. Defaults to None.
          coordmol (list, optional): Coordinate of the molecules in the unit cell. Defaults to None.
          unitcell (list, optional): The lattice parameters of the unitcell. Defaults to None.
          supercell (list, optional): Dimensions of the supercell. Defaults to None.
          unique (int, optional): Number of unique interactions. Defaults to None.
          uniqinter (list, optional): Description of each unique interaction. Defaults to None.
            This list contains six entries. 
            The first two entries describe the type of each molecule in the interaction.
            (Example: [1,1] represents the interaction between similar molecules.)
            The following three are the direction of interaction.
            (Example: [1,0,0] is an interaction in the x-direction)
            The last entry is the direction of the transfer integral.
            (Example: [1] uses the first value of the javg list.)
          javg (list, optional): Transfer integrals for all the pairs in eV. Defaults to None.
          sigma (list, optional): Standard deviation of the transfer integrals of 
            all the pairs in eV. Defaults to None.
          nrepeat (int, optional): Number of iterations with different realizations of 
            disorder. Defaults to None.
          iseed (int, optional): Index of random seed. Defaults to None.
          invtau (int, optional): Characteristic localization time. Defaults to None.
          temp (int, optional): Temperature in eV. Defaults to None.
          lattice_file (json, optional): The structural parameters json file. Defaults to None.
          params_file (json, optional): Parameters for the calculation of mobility. Defaults to None.
          results (dict, optional): Dictionary with resulting localization length 
            and mobilty. Defaults to {}.
      """

      if lattice_file:
         with open(lattice_file + '.json', 'r') as json_file:
            lat_dic = json.load(json_file)
         self.nmuc = lat_dic['nmuc']
         self.coordmol = np.array(lat_dic['coordmol'])
         self.unitcell = np.array(lat_dic['unitcell'])
         self.supercell = lat_dic['supercell']
         self.unique = lat_dic['unique']
         self.uniqinter = np.array(lat_dic['uniqinter'])
      else:
         self.nmuc = nmuc
         self.coordmol = np.array(coordmol)
         self.unitcell = np.array(unitcell)
         self.supercell = supercell
         self.unique = unique
         self.uniqinter = np.array(uniqinter)

      if params_file:
         with open(params_file + '.json', 'r') as json_file:
            par_dic = json.load(json_file)
         self.javg = np.array(par_dic['javg'])
         self.sigma = np.array(par_dic['sigma'])
         self.nrepeat = par_dic['nrepeat']
         self.iseed = par_dic['iseed']
         self.invtau = par_dic['invtau']
         self.temp = par_dic['temp']
      else:
         self.javg = np.array(javg )
         self.sigma = np.array(sigma)
         self.nrepeat = nrepeat
         self.iseed = iseed
         self.invtau = invtau
         self.temp = temp

      self.results = results

      # addind 0 for the case that molecules dont interact
      self.javg = np.insert(self.javg, 0, 0)
      self.sigma = np.insert(self.sigma, 0, 0)

      # making random numbers predictable
      np.random.seed(self.iseed)

   def get_interactions(self):
      """Calculate all the possible interactions in the supercell by translating the unit cell
      interactions and calculating the distance between molecules.

      Returns:
          tuple : (nmol, transinter_mm, distx_mm, disty_mm)
            (Number of molecules in super cell, 
            Translated interaction between a pair of molecules (Matrix size: [nmol, nmol]),
            Distances between all molecules THAT INTERACT with enforced PBC in x-direction,
            Distances between all molecules THAT INTERACT with enforced PBC in y-direction)
      """
      # nmol number of molecules in the supercell
      nmol = self.nmuc * self.supercell[0] * self.supercell[1] * \
         self.supercell[2]
      
      # t number of translations and k directions
      translations_tk = np.mgrid[0:self.supercell[0]:1, 
                                 0:self.supercell[1]:1, 
                                 0:self.supercell[2]:1].reshape(3,-1).T
      transvecs_tk = np.dot(translations_tk, self.unitcell)
      t = len(transvecs_tk)

      # nconnect number of connections
      nconnect = self.unique * t

      # mapping
      mapcell_k = [self.supercell[1] * self.supercell[2], 
                   self.supercell[2], 
                   1]
      map2same_t = self.nmuc * np.dot(translations_tk, mapcell_k)

      # u unique connections to other molecules and enforced PBC
      connec_tuk = (translations_tk[:, None] \
         + self.uniqinter[:, 2:5][None, :])
      connec_tuk[:, :, 0][connec_tuk[:, :, 0] == self.supercell[0]] = 0
      connec_tuk[:, :, 1][connec_tuk[:, :, 1] == self.supercell[1]] = 0
      connec_tuk[:, :, 2][connec_tuk[:, :, 2] == self.supercell[2]] = 0
      map2others_tu = self.nmuc * np.dot(connec_tuk, mapcell_k)
      
      # translated interactions between a pair of molecules
      transinter_mm = np.zeros([nmol, nmol], dtype='int')
      firstmol = (map2same_t[:, None] \
         + self.uniqinter[:, 0][None, :]).reshape(nconnect)
      secondmol = (map2others_tu[:, None] \
         + self.uniqinter[:, 1][None, :]).reshape(nconnect)
      typeinter = np.tile(self.uniqinter[:, 5], t)
      transinter_mm[firstmol - 1, secondmol - 1] = typeinter

      # enforce hermitian
      transinter_mm[secondmol - 1, firstmol - 1] = typeinter

      # translated coordinates for m molecules
      transcoords_mk = (transvecs_tk[:, None] + \
         self.coordmol[None, :]).reshape(nmol, 3)

      # distances between all molecules THAT INTERACT with enforced PBC
      distx_mm = np.zeros([nmol, nmol])
      disty_mm = np.zeros([nmol, nmol])
      for i, j in zip(firstmol, secondmol):
         distx_mm[i-1, j-1] = transcoords_mk[i-1, 0] - transcoords_mk[j-1, 0]
         disty_mm[i-1, j-1] = transcoords_mk[i-1, 1] - transcoords_mk[j-1, 1]

      superlengthx = self.unitcell[0, 0] * self.supercell[0]
      superlengthy = self.unitcell[1, 1] * self.supercell[1]

      distx_mm[distx_mm > superlengthx / 2] -= superlengthx
      distx_mm[distx_mm < -superlengthx / 2] += superlengthx
      disty_mm[disty_mm > superlengthy / 2] -= superlengthy
      disty_mm[disty_mm < -superlengthy / 2] += superlengthy

      return nmol, transinter_mm, distx_mm, disty_mm

   def get_hamiltonian(self, nmol, transinter_mm):
      """Calculating Hamiltonian

      Args:
          nmol (int): Number of molecules in supercell.
          transinter_mm (array): Translated interactions between a pair of molecules 
            (Matrix size: [nmol, nmol])

      Returns:
          array: Hamiltonian
      """
      rnd_mm = np.random.normal(0, 1, size=(nmol, nmol))
      rnd_mm = np.tril(rnd_mm) + np.tril(rnd_mm, -1).T

      hamiltonian_mm = self.javg[transinter_mm] + self.sigma[transinter_mm] * rnd_mm
    
      return hamiltonian_mm

   def get_energies(self, nmol, transinter_mm):
      """Calculate energies by solving Hamiltonian

      Args:
          nmol (int): Number of molecules in supercell.
          transinter_mm (array): Translated interactions between a pair of molecules 
            (Matrix size: [nmol, nmol])

      Returns:
          tuple: (energies_m.real, vectors_mm, hamiltonian_mm)
            (Real part of eigen values, Eigen vectors, Hamiltonian)
      """
      hamiltonian_mm = self.get_hamiltonian(nmol, transinter_mm)

      energies_m, vectors_mm = linalg.eigh(hamiltonian_mm)
      
      return energies_m.real, vectors_mm, hamiltonian_mm

   def get_squared_length(self):
      """Calculate squared transient localization length using
      distances, eigenvectors, eigenenergies and hamiltonian:

      L**2 = (1 / Z) ∑_nm exp(β E_n) |<n | j_x(y) | m>|**2 (2 / ((hbar**2 / tau) + (E_m + E_n)**2))

      <n | j_x(y) | m> = i <n | [H, x(y)] | m>

      Returns:
          tuple: (sqlx, sqly)
            (squared localization length in the x direction, and y direction)
      """
      nmol, transinter_mm, distx_mm, disty_mm = self.get_interactions()

      energies_m, vectors_mm, hamiltonian_mm = self.get_energies(nmol, transinter_mm)

      operatorx_mm = np.matmul(vectors_mm.T, np.matmul(distx_mm * hamiltonian_mm, vectors_mm))
      operatorx_mm -= np.matmul(vectors_mm.T, np.matmul(distx_mm * hamiltonian_mm, vectors_mm)).T

      operatory_mm = np.matmul(vectors_mm.T, np.matmul(disty_mm * hamiltonian_mm, vectors_mm))
      operatory_mm -= np.matmul(vectors_mm.T, np.matmul(disty_mm * hamiltonian_mm, vectors_mm)).T

      partfunc_m = np.exp(energies_m / self.temp)
      partfunc = sum(partfunc_m)

      sqlx = sum(sum(partfunc_m * (operatorx_mm**2 * 2 / (self.invtau**2 + \
         (energies_m[:, None] - energies_m[None, :])**2))))
      sqly = sum(sum(partfunc_m * (operatory_mm**2 * 2 / (self.invtau**2 + \
         (energies_m[:, None] - energies_m[None, :])**2))))

      sqlx /= partfunc
      sqly /= partfunc

      return sqlx, sqly

   def get_disorder_avg_sql(self):
      """Take average of squared localization length considering nrepeat iterations 
      with different disorder realizations

      Returns:
          tuple: (dsqlx, dsqly)
            (disorder-average squared localization length in the x direction, 
            and y direction)
      """
      dsqlx, dsqly = 0, 0
      self.results['squared_length_x'] = []
      self.results['squared_length_y'] = []
      self.results['squared_length'] = []
      self.results['avg_sqlx'] = []
      self.results['avg_sqly'] = []
      self.results['avg_sql'] = []

      for i in tqdm(range(1, self.nrepeat + 1), desc='Calculating average of squared transient localization'):
         sqlx, sqly = self.get_squared_length()
         self.results['squared_length_x'].append(sqlx)
         self.results['squared_length_y'].append(sqly)
         self.results['squared_length'].append((sqlx + sqly) / 2)

         #moving average
         dsqlx -= dsqlx / i
         dsqlx += sqlx / i

         dsqly -= dsqly / i
         dsqly += sqly / i

         self.results['avg_sqlx'].append(dsqlx)
         self.results['avg_sqly'].append(dsqly)
         self.results['avg_sql'].append((dsqlx + dsqly) / 2)

         # print(i, dsqlx, dsqly)

      return dsqlx, dsqly
   
   def get_mobility(self):
      """Calculate charge mobility in the framework of transient localization theory

      Returns:
          tuple: (mobx, moby)
            (mobility in the x direction, and y direction)
      """
      dsqlx, dsqly = self.get_disorder_avg_sql()

      # unit converter unit = ang**2 * e / hbar
      unit = 0.15192674605831966

      # in units of cm**2 / (V s)
      mobx = (1 / self.temp) * self.invtau * dsqlx / 2 * unit
      moby = (1 / self.temp) * self.invtau * dsqly / 2 * unit

      self.results['mobx'] = mobx
      self.results['moby'] = moby

      print('Calculating charge mobility')
      print('mu_x = ', mobx)
      print('mu_y = ', moby)

      return mobx, moby



from ase.io import read
from ase.neighborlist import natural_cutoffs, NeighborList
from scipy import sparse
import numpy as np
import os
from tqdm.auto import trange, tqdm
from halo import Halo
from ase import Atoms
from ase.calculators.gaussian import Gaussian
from elph import chdir, mkdir
import subprocess
from os.path import exists
import json
from collections import OrderedDict

@Halo(text="Reading structure", color='blue', spinner='dots')
def find_structure_file(folder):
    """Searches the current working directory for the molecular structure file. 
        Allowed file types are .cif, .gen, .sdf, or .xyz. 

    Args:
        folder (str): The current working directory.

    Returns:
        str: Molecular structure file. 
    """
    import glob
    
    structure_file = glob.glob(folder + '/*.cif') + \
          glob.glob(folder + '/*.gen') + \
          glob.glob(folder + '/*.sdf') + \
          glob.glob(folder + '/*.xyz')
    print(structure_file)
    structure_file = structure_file[0]

    return structure_file

def get_centers_of_mass(atoms, n_components, component_list):
    """Gets centers of mass for each molecule

    Args:
        atoms (Atoms): Atoms in the system
        n_components (int): Number of molecules in the system
        component_list (dict): maps each atom number to the molecule it belongs to

    Returns:
        list: A list containing the centers of mass for each molecule in the system
    """
    centers_of_mass = []
    for i in range(n_components):
        molIdx_i = i
        molIdxs_i = [ x for x in range(len(component_list)) if component_list[x] == molIdx_i ]
        centers_of_mass.append(atoms[molIdxs_i].get_center_of_mass())
    return centers_of_mass

def compute_total_weight(centers_of_mass):
    """Computes the distances between the centers of mass. We represent the centers of mass as a fully-connected graph with the weight of each edge representing the distance between molecules

    Args:
        centers_of_mass (list): A list containing the centers of mass of each molecule in the system

    Returns:
        tuple: total_weight is a float containing the sum of all weights in the graph. has_dupes is a bool indicating whether there are two molecules overlapping.
    """
    total_weight = 0
    has_dupes = False
    for i in range(len(centers_of_mass)):
        for j in range(i+1, len(centers_of_mass)):
            total_weight +=  np.linalg.norm(centers_of_mass[i] - centers_of_mass[j])
            if np.linalg.norm(centers_of_mass[i] - centers_of_mass[j]) < 0.1:
                has_dupes = True
    return total_weight, has_dupes

def write_structure(label, component_list, molecules, all_atoms):
    """Writes the structure to a file

    Args:
        label (str): The label to be written. Pairs are A, B, C and molecules are 1, 2, 3
        component_list (dict): maps the atom number to the molecule it belongs to
        molecules (list): A list of the molecule numbers to be written
        all_atoms (Atoms): The Atoms object containing the atoms to be written
    """
    atoms = Atoms()
    if isinstance(molecules, list):
        for molecule in molecules:
            idxs = [ i for i in range(len(component_list)) if component_list[i] == molecule ]
            atoms.extend(all_atoms[idxs])
    else:
        idxs = [ i for i in range(len(component_list)) if component_list[i] == molecules ]
        atoms.extend(all_atoms[idxs])
    atoms.set_pbc([False, False, False])
    atoms.set_cell([0, 0, 0])

    mkdir(label)
    atoms.write(label + '/' + label + '.xyz')

@Halo(text="Identifying molecules", color='green', spinner='dots')
def find_neighbors(atoms):
    """Identifies molecules by finding neighboring atoms

    Args:
        atoms (Atoms): Atoms in the system

    Returns:
        tuple: n_components is the number of molecules, component_list is a dict mapping atom number to molecule number, and edges are the a list containing which atoms are neighbors
    """
    neighbor_list = NeighborList(natural_cutoffs(atoms), self_interaction=False, bothways=True)
    neighbor_list.update(atoms)
    matrix = neighbor_list.get_connectivity_matrix(neighbor_list.nl)
    n_components, component_list = sparse.csgraph.connected_components(matrix)
    edges = list(matrix.keys())

    return n_components, component_list, edges

def is_triangle(centers_of_mass):
    """Checks whether centers of mass form a triangle and not a line
    Args:
        centers_of_mass (np.ndarray): An array containing the centers of mass of each molecule
    Returns:
        Bool: Whether the centers of mass form a triangle or not
    """
    l1 = np.linalg.norm(centers_of_mass[1] - centers_of_mass[0])
    l2 = np.linalg.norm(centers_of_mass[2] - centers_of_mass[0])
    l3 = np.linalg.norm(centers_of_mass[2] - centers_of_mass[1])
    return (l1 < l2 + l3) and (l2 < l1 + l3) and (l3 < l1 + l2)

def unwrap_atoms(structure_file=None):
    """Unwraps the molecules and identifies pairs based on a structure file.

    Args:
        structure_file (str, optional): Path to structure file. If none is given, local directory will be automatically searched to find a structure file. Defaults to None.

    Returns:
        str: json string containing the pair definitions based on molecule number.
    """
    folder = os.getcwd()

    if structure_file:
        if os.path.exists(folder + '/../' + structure_file):
            structure_file = folder + '/../' + structure_file
        else:
            structure_file = folder + '/' + structure_file
    else:
        structure_file = find_structure_file(folder)

    atoms = read(structure_file)

    n_components, component_list, edges = find_neighbors(atoms)

    max_bond_len = max(natural_cutoffs(atoms))
    cell = list(atoms.get_cell())
    atoms.set_pbc([False,False,False])
    
    # For each bond, take the lower left and move it upper right until the bond shrinks
    all_positions = atoms.get_positions()
    is_optimized = False
    print("optimizing atoms")
    while not is_optimized:
        is_optimized = True
        for i in range(3):
            for edge in edges:
                positions = all_positions
                distance = np.linalg.norm(positions[edge[0]]-positions[edge[1]])
                if distance > max_bond_len*2:
                    min_pos = positions[edge[0]] if positions[edge[0],i] < positions[edge[1],i] else positions[edge[1]]
                    max_pos = positions[edge[0]] if positions[edge[0],i] >= positions[edge[1],i] else positions[edge[1]]
                    new_pos = min_pos
                    if np.linalg.norm(max_pos - (min_pos + cell[i])) < distance:
                        new_pos = min_pos + cell[i]
                        is_optimized = False
                    if np.array_equal(min_pos,positions[edge[0]]):
                        all_positions[edge[0]] = new_pos
                        all_positions[edge[1]] = max_pos
                    else:
                        all_positions[edge[0]] = max_pos
                        all_positions[edge[1]] = new_pos
    atoms.set_positions(all_positions)
    
    # Construct graph (compute total weight)
    original_centers_of_mass = get_centers_of_mass(atoms, n_components, component_list)
    centers_of_mass = np.copy(original_centers_of_mass)
    weight, has_dupes = compute_total_weight(centers_of_mass)
    test_dirs = cell
    test_dirs.extend([-x for x in test_dirs])

    # Move molecules toward each other
    is_optimized = False
    while not is_optimized:
        is_optimized = True
        for i in range(n_components):
            for j in range(6):
                test_centers_of_mass = np.copy(centers_of_mass)
                test_centers_of_mass[i] += test_dirs[j]
                test_weight, has_dupes = compute_total_weight(test_centers_of_mass)
                if test_weight < weight and not has_dupes:
                    centers_of_mass = np.copy(test_centers_of_mass)
                    weight = test_weight
                    is_optimized = False

    translations = np.zeros([len(atoms), 3])
    for i in range(n_components):
        molIdx = component_list[i]
        molIdxs = [ x for x in range(len(component_list)) if component_list[x] == molIdx ]
        
        dif = centers_of_mass[i] - original_centers_of_mass[i]
        translations[molIdxs,:] = dif

    atoms.translate(translations)
    atoms.center()

     # Construct 3 closest molecules which will form 3 pairs
    original_n_components = n_components

    # If 2 molecules per unit cell, we start with A and B type molecules only
    # If 4 molecules per unit cell, we start with A, B, C, and D molecules
    # For both cases, the high-mobility plane, which we're interested in, consists of A and B molecules only
    # Therefore, we need another copy in both cases
    atoms = atoms * [2,1,1]
    n_components, component_list, edges = find_neighbors(atoms)

    new_atoms = Atoms()
    new_atoms.set_cell(atoms.get_cell())
    counter = 0
    atom_mapping = {}
    for i in range(3):
        # Gives us another A-type molecule for 4-molecule systems
        if i == 2 and original_n_components >= 3:
            idx = original_n_components
        else:
            idx = i
        molIdxs = [ j for j in range(len(component_list)) if component_list[j] == idx ]
        new_atoms.extend(atoms[molIdxs])
        for idx in molIdxs:
            atom_mapping[idx] = counter 
            counter += 1

    # Write atom mapping which will be used later to match phonons to grad J
    with open('atom_mapping.json', 'w') as f:
        f.write(json.dumps(OrderedDict(sorted(atom_mapping.items(), key=lambda t: t[1])), indent=2))

    # Move 3 molecules as close together as possible
    n_components, component_list, edges = find_neighbors(new_atoms)
    coms = get_centers_of_mass(new_atoms, n_components, component_list)
    weight, has_dupes = compute_total_weight(coms)
    
    states = [-3,-2,-1,0,1,2,3]
    best_weight = weight
    best_coms = np.copy(coms)
    same_state_count = 0
    while same_state_count < 3:
        same_state_count += 1
        for molecule_num in range(3):
            for i in states:
                for j in states:
                    for k in states:
                        vec = i*test_dirs[0] + j*test_dirs[1] + k*test_dirs[2]
                        test_coms = np.copy(coms)
                        test_coms[molecule_num] += vec
                        weight, has_dupes = compute_total_weight(test_coms)
                        if not has_dupes and is_triangle(test_coms):
                            if weight < best_weight:
                                best_weight = weight
                                best_coms = np.copy(test_coms)
                                same_state_count = 0

    translations = np.zeros([len(new_atoms), 3])
    displacements = best_coms - coms
    for i in range(3):
        molIdxs = [ j for j in range(len(component_list)) if component_list[j] == i ]
        translations[molIdxs] = displacements[i]
    new_atoms.translate(translations)

    final_atoms = Atoms()
    for i in range(3):
        molIdxs = [ j for j in range(len(component_list)) if component_list[j] == i ]
        final_atoms.extend(new_atoms[molIdxs])
        
    final_atoms.center()
    final_atoms.set_pbc([False, False, False])
    final_atoms.set_cell([0, 0, 0])
    n_components, component_list, edges = find_neighbors(final_atoms)
    final_atoms.write('all_pairs.xyz')


    # Create structures with each pair of atoms
    """
    Directory structure will be as follows:
    >main_folder
        >1
            -1.com
            ~1.log
            ~fort.7 > 1.pun
            >Displacements
                -0
                -1
                ...
        >2
            -2.com
            ~2.log
            ~fort.7 > 2.pun
        >3
            -3.com
            ~3.log
            ~fort.7 > 3.pun
        >A
            -PairA.com
            ~PairA.log
            ~fort.7 > PairA.pun
        >B
            -PairB.com
            ~PairB.log
            ~fort.7 > PairB.pun
        >C
            -PairC.com
            ~PairC.log
            ~fort.7 > PairC.pun
    """

    molecules = {'1':0, 
                 '2':1, 
                 '3':2}

    for key, value in molecules.items():
        write_structure(key, component_list, value, final_atoms)

    pairs = {'A':[0, 1], 
             'B':[1, 2],
             'C':[0, 2]}

    for key, value in pairs.items():
        cycle = value
        write_structure(key, component_list, cycle, final_atoms)

    with open('all_pairs.json', 'w', encoding='utf-8') as f:
        json.dump(pairs, f, ensure_ascii=False, indent=4)

    return pairs

@Halo(text="Running Gaussian calculation", color='red', spinner='dots')
def get_orbitals(atoms, name):
    """Runs gaussian to compute the orbitals for a system

    Args:
        atoms (Atoms): Atoms in the system
        name (str): The name of the system
    """
    if not exists(name + '.pun'):
        atoms.calc = Gaussian(mem='4GB',
                              nprocshared=12,
                              label=name,
                              save=None,
                              method='b3lyp',
                              basis='3-21G*',
                              scf='tight',
                              pop='full',
                              extra='nosymm punch=mo iop(3/33=1)')
        atoms.get_potential_energy()
        os.rename('fort.7', name + '.pun')
    print(['Simulation {} is done' .format(name)])

@Halo(text="Calculating transfer integral", color='red', spinner='dots')
def catnip(path1, path2, path3):
    """Runs Catnip to determine the transfer integral

    Args:
        path1 (str): Path to the first Gaussian Result
        path2 (str): Path to the second Gaussian Result
        path3 (str): Path to the third Gaussian Result

    Returns:
        str: the transfer integral for the system
    """
    path1 += '.pun'
    path2 += '.pun'
    path3 += '.pun'
    cmd = os.environ['ELPH_CATNIP_CMD']
    cmd += " -p_1 {} -p_2 {} -p_P {}" .format(path1, path2, path3)
    output = subprocess.check_output(cmd, shell=True)
    return output.decode('ascii').split()[-13]

def get_javerage(pair):
    """Computes the transfer integral for a pair of molecules

    Args:
        pair (list): Contains the molecule numbers in the pair

    Returns:
        str: The transfer itegral
    """
    paths = []

    # Gaussian run for each molecule
    for mol in pair[1]:
        name = str(mol + 1)
        with chdir(name):
            atoms = read(name + '.xyz')
            paths.append(name + '/' + name)
            get_orbitals(atoms, name)
    
    # Gaussian run for the pair
    with chdir(pair[0]):
        atoms = read(pair[0] + '.xyz')
        paths.append(pair[0] + '/' + pair[0])
        get_orbitals(atoms, pair[0])

    # Calculate J 
    j = catnip(paths[0], paths[1], paths[2])

    print('J_{} = {}' .format(pair[0], j))
    return j

def javerage():
    """main function, calls unwrap atoms and get_javerage for each pair in the system.
    """
    pairs = unwrap_atoms()

    for pair in pairs.items():
        j = get_javerage(pair)
        data = {pair[0]: j}
        with open('J_' + pair[0] + '.json', 'w', encoding='utf-8') as f:
             json.dump(data, f, ensure_ascii=False, indent=4)

if __name__ == '__main__':
    javerage()


import json
from pathlib import Path
from elph.javerage import javerage
from elph.sigma import sigma
from elph.visualization import view
from elph.molecules import Molecules

def write_lattice_file():
   """Write the lattice parameters json file
   """
   lattice = {'nmuc':2,
              'coordmol':[[0.0, 0.0, 0.0], [0.5, 0.5, 0.0]],
              'unitcell':[[1.0, 0.0, 0.0], [0.0, 1.7321, 0.0], [0.0, 0.0, 1000.0]],
              'supercell':[16, 16, 1],
              'unique':6,
              'uniqinter':[[1, 1, 1, 0, 0, 1], 
              [2, 2, 1, 0, 0, 1], 
              [1, 2, 0, 0, 0, 3], 
              [2, 1, 1, 0, 0, 2], 
              [2, 1, 0, 1, 0, 2], 
              [2, 1, 1, 1, 0, 3]]
   }
   with open('lattice.json', 'w', encoding='utf-8') as f:
      json.dump(lattice, f, ensure_ascii=False, indent=4)

def write_params_file():
   params = {'javg':[0.058, 0.058, 0.058],
             'sigma':[0.029, 0.029, 0.029],
             'nrepeat':50,
             "iseed":3987187,
             'invtau':0.005,
             'temp':0.025
   }
   with open('params.json', 'w', encoding='utf-8') as f:
      json.dump(params, f, ensure_ascii=False, indent=4)

def main(args=None):
   import argparse

   description = "Transient Localization Theory command line interface"

   example_text = """
   examples:

   Calculate charge mobility with:
      elph --mobility
   """

   formatter = argparse.RawDescriptionHelpFormatter
   parser = argparse.ArgumentParser(description=description,
                                    epilog=example_text, 
                                    formatter_class=formatter)

   help = """
   All calculations require a lattice JSON 
   file with the following properties:

   lattice.json:
      nmuc: 
      coordmol: 
      unitcell: 
      supercell: 
      unique:
      uniqinter: 

   """
   parser.add_argument('--lattice_file', nargs='*', help=help,
                        default='lattice', type=str)

   help = """
   All calculations require a params json 
   file with the following properties:

   params.json:
      javg: 
      sigma: 
      nrepeat:
      iseed: 
      invtau:
      temp:

   """
   parser.add_argument('--params_file', nargs='*', help=help,
                        default='params', type=str)

   help = ("write example of lattice and params files")
   parser.add_argument('--write_files', action='store_true' , help=help)

   help = """
   Calculate average transfer integral 
   for all unique pairs of molecules

   Gaussian calculator is required. Define the path with:
   ASE_GAUSSIAN_COMMAND='g09 < PREFIX.com > PREFIX.log'

   """
   parser.add_argument('--javerage', action='store_true' , help=help)

   help = """
   Calculate variance of the transfer integral 
   for all unique pairs of molecules

   Javerage step and phonon modes (mesh.yaml) are required

   """
   parser.add_argument('--sigma', action='store_true' , help=help)

   help = """
   Visualization tool of sigma

   [atoms] View the total sigma contribution for each atom
   [modes] [[n]] View the n phonon modes with the highest sigma contribution

   """
   parser.add_argument('--view', nargs='*', type=str, help=help)

   help = ("Calculate charge mobility")
   parser.add_argument('--mobility', action='store_true' , help=help)

   args = parser.parse_args(args)

   if args.write_files:
      write_lattice_file()
      write_params_file()
      return

   print('Initializing ElPh')

   if args.javerage:
      javerage()

   if args.sigma:
      sigma()

   if args.view:
      view(*args.view)

   if args.mobility:
      if not Path(args.lattice_file + '.json').is_file():
         msg = 'Lattice file could not be found'
         raise FileNotFoundError(msg)

      if not Path(args.params_file + '.json').is_file():
         msg = 'Params file could not be found'
         raise FileNotFoundError(msg)

      mols = Molecules(lattice_file=args.lattice_file, 
      params_file=args.params_file)
         
      mobx, moby = mols.get_mobility()
      with open('results.json', 'w', encoding='utf-8') as f:
         json.dump(mols.results, f)

if __name__  == '__main__':
   main()



import numpy as np
import json
from elph.sigma import load_phonons, get_dj_matrix

def heat_atoms(molpair, sigma_eav):
    """Scatter plot of atoms with heat indicating 
    the highest contributions to sigma. Save plot data 
    (atomic positions, atomic sizes and sigma contribution) 
    to numpy compressed file.

    Args:
        molpair (str): Molecular pair label
        sigma_eav (array): Sigma array of sizes modes, atoms and directions
    """
    from ase.io import read
    import matplotlib.pyplot as plt
    # import plotly.graph_objects as go

    atoms = read(molpair + '/' + molpair + '.xyz')
    pos = atoms.get_positions()
    masses = atoms.get_masses()
    x, y, z = pos[:, 0], pos[:, 1], pos[:, 2]
    sigma = np.sum(np.sum(sigma_eav, axis=-1), axis=0)
    
    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(projection='3d')
    s = ax.scatter(x, y, z, s = 20 * masses, c=sigma)
    fig.colorbar(s)
    plt.show()

    data = {'x': x,
            'y': y,
            'z': z,
            'size': 20 * masses,
            'sigma': sigma}

    np.savez_compressed('view_atoms_' + molpair + '.npz', **data)

    # layout = go.Layout(
    #          scene=dict(
    #              aspectmode='data'
    #         ))
    # fig = go.Figure(data=[go.Scatter3d(x=x, 
    #                                    y=y, 
    #                                    z=z, 
    #                                    mode='markers',
    #                                    marker=dict(size=10*np.sqrt(masses/np.pi),
    #                                                color=sigma,
    #                                                colorscale='Viridis',
    #                                                opacity=0.8,
    #                                                ))],
    #                 layout=layout)
    # fig.write_html(molpair + '.html')
    # fig.show()

def heat_modes(molpair, sigma_eav, vecs_eav, n):
    """Scatter plot of atoms with arrows indicating
    phonon displacements of n phonons modes with the 
    highest sigma contributions. Save plot data 
    (atomic positions, vector directions, atomic sizes 
    and sigma contribution) to numpy compressed file.

    Args:
        molpair (str): Molecular pair label
        sigma_eav (array): Sigma array of sizes modes, atoms and directions
        vecs_eav (array): Array of phonon eigendisplacements of size modes, atoms and directions
        n (int): Top n phonon modes with highest sigma contribution
    """
    from ase.io import read
    import matplotlib.pyplot as plt

    atoms = read(molpair + '/' + molpair + '.xyz')
    pos = atoms.get_positions()
    masses = atoms.get_masses()
    x, y, z = pos[:, 0], pos[:, 1], pos[:, 2]
    sigma = np.sum(np.sum(sigma_eav, axis=-1), axis=-1)

    ind = np.argsort(sigma)[-n:][::-1]

    data = {'x': x,
            'y': y,
            'z': z,
            'size': 20 * masses}

    for i in ind:
        fig = plt.figure(figsize=(8, 8))
        ax = fig.add_subplot(projection='3d')
        u = vecs_eav[i, :, 0]
        v = vecs_eav[i, :, 1]
        w = vecs_eav[i, :, 2]
        data['u'] = u
        data['v'] = v
        data['w'] = w
        ax.quiver(x, y, z, u, v, w, length=2, normalize=True)
        ax.scatter(x, y, z, s = 20 * masses)
        plt.show()

        np.savez_compressed('view_modes_' + molpair + '_' + '{}' .format(i) + '_' + '.npz', **data)

def sigma_contribution(pair_atoms, dj_av, temp):
    """Standard deviation (sigma) in function of
    modes, atoms and directions.

    Args:
        pair_atoms (list): List containing atom indices of molecular pair
        dj_av (array): Delta of J in function of atoms and directions
        temp (float): Temperature in eV

    Returns:
        typle: sigma and eigendisplacements in function of modes, atoms and directions
    """
    freqs_e, vecs_eav, _ = load_phonons(pair_atoms)
    epcoup_eav = vecs_eav * dj_av[None, :, :]
    ssigma_eav = epcoup_eav**2 / (2 * np.tanh(freqs_e[:, None, None] / (2 * temp)))

    return np.sqrt(ssigma_eav), vecs_eav

def get_sigma(pair, delta, temp, mode, n):
    """Read transfer integrals of displacements systems and calculate sigma

    Args:
        pair (dic): Dictionary containing molecular indices of pair
        delta (float): Step size to displace atoms in Ang
        temp (float): Temperature in eV
        mode (str): Mode of visualization (atoms or modes)
        n (int): Top n phonon modes with highest sigma contribution

    Raises:
        NotImplementedError: Error raised if the chosen mode is not atoms or modes
    """
    mol1 = str(int(pair[1][0]) + 1)
    mol2 = str(int(pair[1][1]) + 1)
    molpair = pair[0]

    jlists = np.load(molpair + '_disp_js.npz')['js']
    dj_matrix_av = get_dj_matrix(jlists, delta)
    offset = len(dj_matrix_av) // 2
    pair_atoms = np.concatenate([np.arange((int(mol1) - 1) * offset, 
                                            int(mol1) * offset), 
                                 np.arange((int(mol2) - 1) * offset, 
                                            int(mol2) * offset)])

    sigma_eav, vecs_eav = sigma_contribution(pair_atoms, dj_matrix_av, temp)

    if mode == 'atoms':
        heat_atoms(molpair, sigma_eav)
    elif mode == 'modes':
        heat_modes(molpair, sigma_eav, vecs_eav, n)
    else:
        msg = 'The available visualization results are atoms and modes'
        raise NotImplementedError(msg)

def view(mode='atoms', n=3):
    """Main function of the visualization module. There are two modes:

    [atoms]: Scatter plot of atoms with heat indicating 
    the highest contributions to sigma.

    [modes] [[n]] : Scatter plot of atoms with arrows indicating
    phonon displacements of n phonons modes with the 
    highest sigma contributions

    Args:
        mode (str, optional): Mode of visualization (atoms or modes). Defaults to 'atoms'.
        n (int, optional): Top n phonon modes with highest sigma contribution. Only useful for modes. Defaults to 3.
    """
    with open('all_pairs.json', 'r') as json_file:
        pairs = json.load(json_file)
    
    for pair in pairs.items():
        get_sigma(pair, delta=0.01, temp=0.025, mode=mode, n=int(n))

if __name__ == '__main__':
    view()


import numpy as np
from ase.io import read
from shutil import copyfile
from elph.javerage import get_orbitals, catnip
from elph.phonons import write_phonons
from elph import chdir, mkdir
import json
import os

def load_phonons(pair_atoms, phonon_file='phonon.npz', 
                 map_file='atom_mapping.json'):
    """Loads phonon modes and returns frequencies, eigenvectors and number of q points

    Args:
        pair_atoms (list): Atomic indices of the interacting pair
        phonon_file (str, optional): Numpy file with the phonon modes. Defaults to 'phonon.npz'
        map_file (str, optional): JSON file generate to map atomic indices 
        of the wraped to the unwraped structure. Defaults to 'atom_mapping.json'

    Returns:
        tuple: vector of frequencies, vector of eigenvectors, and integer of the number of qpoints
    """
    # read phonon modes file
    phonon = np.load(phonon_file)

    # read mapping file
    with open(map_file, 'r') as json_file:
        mapping = list(map(int, json.load(json_file).keys()))
    
    # use mapping to order the wrapped phonon modes
    # based on the unwrapped atoms
    if len(mapping) > len(phonon['vecs'][0]):
        vecs_eav = np.tile(phonon['vecs'], [1, 2, 1])[:, mapping, :]
    else:
        vecs_eav = phonon['vecs'][:, mapping, :]

    # selecting only the phonon modes relevant to the 
    # interaction pair of molecules
    vecs_eav = vecs_eav[:, pair_atoms, :]

    return phonon['freqs'], vecs_eav, phonon['nq']

def get_displacements(atoms):
    """Returns displacement of each atom in each direction

    Args:
        atoms (Atoms): Atoms objects

    Yields:
        yield: (atom number, direction, sign)
    """
    latoms = len(atoms)
    for ia in range(latoms):
        for iv in range(3):
            for sign in [-1, 1]:
                yield (ia, iv, sign)

def displace_atom(atoms, ia, iv, sign, delta):
    """Displace one atomic position in the Atoms object

    Args:
        atoms (Atoms): Atoms object
        ia (int): atom index
        iv (int): direction index
        sign (str): sign of displacement in each direction
        delta (float): size of displacement

    Returns:
        Atoms: Updated Atoms object
    """
    new_atoms = atoms.copy()
    pos_av = new_atoms.get_positions()
    pos_av[ia, iv] += sign * delta
    new_atoms.set_positions(pos_av)
    return new_atoms

def finite_dif(delta=0.01):
    """Compute Gaussian calculation for displaced system

    Args:
        delta (float, optional): Size of displacement. Defaults to 0.01.
    """
    atoms = read('static.xyz')
    for ia, iv, sign in get_displacements(atoms):
        prefix = 'dj-{}-{}{}{}' .format(int(delta * 1000), 
                                        ia,
                                        'xyz'[iv],
                                        ' +-'[sign])
        new_structure = displace_atom(atoms, ia, iv, sign, delta)
        get_orbitals(new_structure, prefix)

def get_dj_matrix(jlists, delta):
    """Matrix containing gradient of J, the transfer integral

    Args:
        jlists (list): list of transfer integrals
        delta (float): size of displacement

    Returns:
        array: matrix containing gradient of j
    """
    latoms = len(jlists) // 6
    # array with j - delta (j minus)
    jm = np.empty([latoms, 3])
    jm[:, 0] = jlists[0::6]
    jm[:, 1] = jlists[2::6]
    jm[:, 2] = jlists[4::6]
    
    # array with j + delta (j plus)
    jp = np.empty([latoms, 3])
    jp[:, 0] = jlists[1::6]
    jp[:, 1] = jlists[3::6]
    jp[:, 2] = jlists[5::6]

    dj_matrix = (np.abs(jp) - np.abs(jm)) / (2 * delta)

    return dj_matrix

def get_deviation(pair_atoms, dj_av, temp):
    """Calculate standard deviation (sigma) of the transfer integral

    Args:
        pair_atoms (array): Numpy array of atomic indices of interacting pairs
        dj_av (array): Numpy array of gradient of J for all atoms in the pair for all directions
        temp (float): Temperature in eV

    Returns:
        float: standard deviation (sigma) of the transfer integral
    """
    freqs_e, vecs_eav, nq = load_phonons(pair_atoms)
    epcoup_e = np.einsum('av,eav->e', dj_av, vecs_eav)
    ssigma = (1 / nq) * np.sum(epcoup_e**2 / \
        (2 * np.tanh(freqs_e / (2 * temp))))

    return np.sqrt(ssigma)

def get_sigma(pair, delta=0.01, temp=0.025):
    """Calculate standard deviation of the transfer integral between a molecule pair using finite differences

    Args:
        pair (tuple): Pair name and molecules indices
        delta (float, optional): Size of displacement. Defaults to 0.01.
        temp (float, optional): Temperature in eV. Defaults to 0.025.

    Returns:
        float: standard deviation of the transfer integral in eV
    """
    mol1 = str(pair[1][0] + 1)
    mol2 = str(pair[1][1] + 1)
    molpair = str(pair[0])
    if not os.path.exists(molpair + '_disp_js.npz'):
        jlists = []
        # run Gaussian for displacements of first molecule
        with chdir(mol1):
            mkdir('displacements')
            with chdir('displacements'):
                copyfile('../' + mol1 + '.xyz', 'static.xyz')
                finite_dif(delta)

        # run Gaussian for displacements of second molecule
        with chdir(mol2):
            mkdir('displacements')
            with chdir('displacements'):
                copyfile('../' + mol2 + '.xyz', 'static.xyz')
                finite_dif(delta)

        # run Gaussian for displacements of the pair
        with chdir(molpair):
            mkdir('displacements')
            with chdir('displacements'):
                copyfile('../' + molpair + '.xyz', 'static.xyz')
                finite_dif(delta)

        # calculating j for each displacement of the first molecule
        path1 = mol1 + '/displacements/'
        path2 = mol2 + '/' + mol2
        path3 = molpair + '/displacements/'

        atoms = read(path1 + 'static.xyz')
        offset = len(atoms)
        for ia, iv, sign in get_displacements(atoms):
            prefix = 'dj-{}-{}{}{}' .format(int(delta * 1000), 
                                                ia,
                                                'xyz'[iv],
                                                ' +-'[sign])
            j = catnip(path1 + prefix, path2, path3 + prefix)
            jlists.append(j)

        # calculating j for each displacement of the second molecule
        path1 = mol1 + '/' + mol1
        path2 = mol2 + '/displacements/'
        path3 = molpair + '/displacements/'

        atoms = read(path2 + 'static.xyz')
        for ia, iv, sign in get_displacements(atoms):
            prefix_mol = 'dj-{}-{}{}{}' .format(int(delta * 1000), 
                                                    ia,
                                                    'xyz'[iv],
                                                    ' +-'[sign])
            prefix_pair = 'dj-{}-{}{}{}' .format(int(delta * 1000), 
                                                     ia + offset,
                                                    'xyz'[iv],
                                                    ' +-'[sign])
            j = catnip(path1, path2 + prefix_mol, path3 + prefix_pair)
            jlists.append(j)

        data = {'js': jlists}   
        np.savez_compressed(molpair + '_disp_js.npz', **data)

    else:
        jlists = np.load(molpair + '_disp_js.npz')['js']

    # Create GradJ matrix with a atoms and v directions
    dj_matrix_av = get_dj_matrix(jlists, delta)

    # Calculate sigma
    offset = len(dj_matrix_av) // 2
    pair_atoms = np.concatenate([np.arange((int(mol1) - 1) * offset, 
                                            int(mol1) * offset), 
                                 np.arange((int(mol2) - 1) * offset, 
                                            int(mol2) * offset)])

    sigma = get_deviation(pair_atoms, dj_matrix_av, temp)
    data = {molpair: sigma}
    
    print('Sigma_{} = {}' .format(pair[0], sigma))

    return sigma

def sigma():
    """Write phonon modes from phonopy result, 
    and calculate the standard deviation (sigma)
    for each pair of molecules. Write sigmas to JSON files
    """
    write_phonons()

    with open('all_pairs.json', 'r') as json_file:
        pairs = json.load(json_file)
    
    for pair in pairs.items():
        sigma = get_sigma(pair, delta=0.01)
        data = {pair[0]: sigma}
        with open('Sigma_' + pair[0] + '.json', 'w', encoding='utf-8') as f:
             json.dump(data, f, ensure_ascii=False, indent=4)

if __name__ == '__main__':
    sigma()


__version__ = '1.0'
name = 'elph'

import os
import errno
from contextlib import contextmanager

def mkdir(folder):
    """Creates folder in current wd unless OSError occurs.

    Args:
        folder (str): Name of folder to create.
    """
    try:
        os.mkdir(folder)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise

@contextmanager
def chdir(folder):
    """Changes the working directory to folder if not already current wd. 

    Args:
        folder (str): Name of folder to make wd.
    """
    dir = os.getcwd()
    os.chdir(str(folder))
    yield
    os.chdir(dir)



import phonopy
import numpy as np

def write_phonons(mesh=[8, 8, 8], phonopy_file="phonopy_params.yaml"):
    """Calculate phonon modes based on the force constants from phonopy and write 
    phonon frequencies, displacements, and number of q points to numpy file.
    It requires the FORCE_SETS file.

    Args:
        mesh (list, optional): Sampling of the q space. Defaults to [8, 8, 8].
        phonopy_file (str, optional): Params file of phonopy. Defaults to "phonopy_params.yaml".
    """
    phonon = phonopy.load(phonopy_file)
    phonon.run_mesh(mesh, with_eigenvectors=True)
    
    # nqpoints x nbands from phonopy to e modes
    freqs_e = phonon.mesh.frequencies.flatten()

    # converting energy unit
    thz2ev = 4.13566733e-3 
    freqs_e *= thz2ev # eV

    # nqpoints x (natoms x 3 directions) x nbands from phonopy
    vecs = phonon.mesh.eigenvectors

    # masses, frac_coords and qpoints
    masses_a = phonon.mesh._cell.masses
    fcoords_av = phonon.mesh._cell.get_scaled_positions()
    qpoints_qv = phonon.mesh.qpoints
    nq = len(qpoints_qv)

    # transform eigenvectors to eigendisplacements (phonopy modulations)
    factor_qa = (np.exp(2j * np.pi * np.dot(fcoords_av, qpoints_qv.T)) / np.sqrt(masses_a)[:, None]).T
    vecs = np.repeat(factor_qa, 3).reshape(nq, -1)[:, :, None] * vecs
    vecs /= np.sqrt(len(masses_a))
        
    # e modes, a atoms, v directions 
    vecs = np.transpose(vecs, axes=[0, 2, 1])
    vecs_eav = vecs.reshape(len(freqs_e), -1, 3)

    # add phase factor (phonopy modulations)
    vecs_e = vecs_eav.reshape(len(vecs_eav), -1)
    index_maxarr = np.abs(vecs_e).argmax(1)
    maxarr = np.take_along_axis(vecs_e, index_maxarr[:, None], axis=1)[:, 0]
    phase_for_zero = maxarr / np.abs(maxarr)
    phase_factor = np.exp(1j * np.pi / 180) / phase_for_zero
    vecs_eav *= phase_factor[:, None, None]

    # avoid negative (imaginary) frequencies
    ind = np.where(freqs_e > 0)
    freqs_e = freqs_e[ind]
    vecs_eav = vecs_eav[ind]

    data = {'freqs': freqs_e,
            'vecs': vecs_eav.real,
            'nq': nq}
    np.savez_compressed('phonon.npz', **data)


import re
from pathlib import Path
from setuptools import setup, find_packages

txt = Path('elph/__init__.py').read_text()
version = re.search("__version__ = '(.*)'", txt).group(1)

long_description = Path('README.md').read_text()

setup(
    name='elph',
    version=version,
    description='Electron Phonon coupling within the Transient Localization Theory',
    #long_description=long_description,
    #long_description_content_type='text/markdown',
    author='Lucas Cavalcante',
    author_email='lsrcavalcante@ucdavis.edu',
    url='https://github.com/lucassamir/ElPh',
    packages=find_packages(),
    include_package_data=True,
    install_requires=['numpy', 'scipy', 'ase>=3.20.0', 'tqdm', 'halo', 'phonopy', 'sphinx_rtd_theme'],
    extras_require={'docs': ['sphinx', 'sphinxcontrib-programoutput']},
    entry_points='''
        [console_scripts]
        elph=elph.main:main
    ''',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Environment :: Console',
        'License :: OSI Approved :: '
        'GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: Unix',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Topic :: Scientific/Engineering :: Physics'
    ])
